import holoviews as hv
import numpy as np
from immutabledict import immutabledict
from bokeh.models import HoverTool
from straxen.bokeh_utils import _patches_x_y, peak_tool_tip


# noinspection PyArgumentList
class PlotPeakLikeData:
    """Base class to plot peak like data for any type of detector.
    """
    
    # Properties to be overwritten by sub-clases if needed.
    _xlabel='x-label'  # Label for x axis defined in _set_labels
    _ylabel='y-label'  # Label for y axis defined in _set_labels
    _label='label'  # Label for plot defined in _set_labels
    _scaler=1
    keep_amplitude_per_sample=False
    _vdim_labels = immutabledict()
    _never_include_fields = ()
    
    def __init__(self,
                 opts=immutabledict(),
                ):
        """
        :param opts_dict: Dictionary of options which will be applied to 
            hv.Area and hv.Curve plots. If you want to change a setting
            which is not common by both plot types please specify them
            in "plot_peak" or "plot_peaks".
        """
        if not isinstance(opts, (dict, immutabledict)):
            raise ValueError('"opts_dict" must be a dictionary!')
        self.opts_dict = opts
        
    def plot_peak(self, peak):
        """User defined function for plotting a single peak with correct
        labels etc.
        """
        raise NotImplementedError()
        
    def plot_peaks(self, peaks):
        """User defined function for plotting multiple peaks into the
        same figure.
        """
        raise NotImplementedError()   
        
    def _set_labels(self, peaks, peak_type, time_prefix):
        """User defined function which should set the xlabel, ylabel and
        plot label for our peak plots. Also should define the scaler 
        which scales the time axis in desired unit e.g. µs or ns.
        """
        raise NotImplementedError()   
    
    
    def _plot_peak(self, 
                   peak, 
                   label,
                   group_label='Peak',
                   opts_curve=immutabledict(), 
                   opts_area=immutabledict(),
                   _relative_start_time=0,
                  ):
        """Plots area and curve to display a peak. 
        :param peak: Peak to be displayed
        :param label: Label to be used in legend.
        :param opts_curve: Additional options to be applied to hv.Curve
        :param opts_area: Additional options to be applied to hv.Area
        :param _relative_start_time: Time to which the peak should be 
            plotted relatively. E.g. event start time.
        
        :returns: hv.Area and hv.Curve
        """
        data = self.get_peak_data(peak, 
                                  relative_start=_relative_start_time,
                                 )
        
        area = hv.Area(data,
                      kdims=hv.Dimension('xs', label=self._xlabel), 
                      vdims=[hv.Dimension('ys', label=self._ylabel),],
                      label=label,
                      group=group_label,
                     ).opts(alpha=0.3, **opts_area, framewise=True)
        
        curve = hv.Curve(data,
                        kdims=hv.Dimension('xs', label=self._xlabel), 
                        vdims=[hv.Dimension('ys', label=self._ylabel),] + self._vdims,
                        label=label,
                        group=group_label,
                       ).opts(**opts_curve, framewise=True)
        
        return area, curve
    
    def _init_peak_plot(self, peak, peak_type, time_prefix):
        """Initalizes empty plots for plotting single or muliple-peaks 
        into the same figure. Intializes plot legend labels as well as
        other dimensions.
        """
        self._set_single_valued_dimensions(peak)
        self._set_labels(peak, peak_type, time_prefix)
        
        area=hv.Area(None,
            kdims=hv.Dimension('xs', label=self._xlabel), 
            vdims=[hv.Dimension('ys', label=self._ylabel),],
            ).opts(framewise=True)

        curve=hv.Curve(None,
                      kdims=hv.Dimension('xs', label=self._xlabel), 
                      vdims=[hv.Dimension('ys', label=self._ylabel),] + self._vdims,
                     ).opts(framewise=True)
        return area, curve
    
    
    def _set_single_valued_dimensions(self, peak):
        """Extracts information from peaks. Defines a list of single 
        valued dimensions which will be extracted for the plot in 
        _get_peak_data.

        Defines holovies vdim Dimensions for a proper hover-tool 
        display. Looks up unit describition in _vdim_labels which have
        to be specified by the user.
        """
        self.single_dimension_fields = self._get_single_dimension_fields(peak)
        vdims = []

        for field_name in self.single_dimension_fields:
            if field_name in self._never_include_fields:
                continue
            vdims.append(hv.Dimension(field_name, 
                                      label=self._vdim_labels.get(field_name, field_name)))
        self._vdims = vdims        

    @staticmethod
    def _get_single_dimension_fields(peak):
        """Function which returns all single valued qunatity names for 
        a given peak.
        """
        field_names = []
        for field_name in peak.dtype.names:
            value = peak[field_name]
            _is_single_value_quantity = value.ndim == 1
            if _is_single_value_quantity:
                field_names.append(field_name)
        
        return field_names
    
    def _is_time_based_quantity(self, field_name):
        """Function which checks if specified parameter is a quantiy 
        which is based on time, like range, width or time properties. 
        """
        raise NotImplementedError()
    
    def get_peak_data(self, peak, relative_start=0):
        """Function which extracts base peaks information as a 
        dictionary. Can be further customized in sub-classes if needed.

        :param peak: Peaks to be plotted.
        :param relative_start: t0 from which on the peaks should be plotted.
        :return: dictionary
        """
        # Wrapper in case one hast to customized things further in one
        # of the sub-clases.
        return self._get_peak_data(peak, 
                                   relative_start=relative_start)

    def _get_peak_data(self, peak, relative_start=0):
        """
        Function which extracts base peaks information as a dictionary.
        Extracts all information which are stored as single valued 
        parameters.

        :param peaks: Peaks to be plotted.
        :param relative_start: t0 from which on the peaks should be plotted.
        :return: dictionary
        """
        x, y = self._patches_x_y(peak)
        x -= relative_start  # relative to first peak
        x = x / self._scaler # Scaler to convert ns to µs or other units
        
        # Coordinates for peak patches
        data={'xs': x,  
              'ys': y,
             }
        
        # Add other single valued fields which we can find to our 
        # dictionary. Adjust time parameters according to scaling.
        for field_name in  self.single_dimension_fields:
            value = peak[field_name]
            
            if self._is_time_based_quantity(field_name):
                data[field_name] = value/self._scaler
            else:
                data[field_name] = value
        
        return data
    
    
    def _patches_x_y(self, peak):
        """
        Creates x,y coordinates needed to make a stepwise function with
        hv.Areas.
        
        :param peak: Peak for which we need the x/y samples
        :returns: Tuple of x, y
        """
        if self.keep_amplitude_per_sample:
            dt_a = 1
        else:
            dt_a = peak['dt']

        x = np.arange(peak["length"])
        xx = np.zeros(2 * len(x), dtype=x.dtype)
        mx = 0.5 * (x[1::] + x[:-1])
        xx[1:-1:2] = mx
        xx[2::2] = mx
        xx[0] = 1.5 * x[0] - 0.5 * x[1]
        xx[-1] = 1.5 * x[-1] - 0.5 * x[-2]
        xx = np.array(xx) * peak['dt'] + peak['time']
        y = peak['data'][:peak['length']]
        yy = np.zeros(2 * len(x))
        yy[0::2] = y
        yy[1::2] = y
        yy = np.array(yy) / dt_a

        # baseline since we'll fill underneath
        xx = np.concatenate([[xx[0]], xx, [xx[-1]]])
        yy = np.concatenate([[0], yy, [0]])
        return xx, yy


# noinspection PyArgumentList
class PlotPeaksTPC(PlotPeakLikeData):
    """Class which plots single or multiple peaks for interactive 
    holoviews display.
    """
    time_in_us=False
    _never_include_fields = ('time', 
                             'endtime',
                             'center_time',
                             'channel', 
                             'type', 
                             'max_goodness_of_split',
                            )
    # Use round brakets for units which should not be affected by 
    # "time_in_µs".
    _vdim_labels = {
        'time': 'time (ns)',  
        'endtime': 'endtime (ns)',
        'center_time': 'center_time (ns)',
        'rise_time': 'rise_time [ns]',
        'range_50p_area': 'range_50p_area [ns]',
        'range_90p_area': 'range_90p_area [ns]',
        'area': 'area [PE]',
        'max_pmt_area': 'max_pmt_area [PE]',
    }
    
    def __init__(self, 
                 opts=immutabledict(),
                 keep_amplitude_per_sample=False,
                ):
        super().__init__(opts)
        self.keep_amplitude_per_sample = keep_amplitude_per_sample
        
    def _set_labels(self, peak, peak_type, time_prefix):
        if self.time_in_us:
            scaler = 1000
            xlabel = f'{time_prefix} Time [µs]'
            self._change_units()
        else:
            xlabel = f'{time_prefix} Time [ns]'
            scaler = 1
        
        if self.keep_amplitude_per_sample:
            ylabel = f'{peak_type} Amplitude [pe/sample]'
        else:
            ylabel = f'{peak_type} Amplitude [pe/ns]'
            
        self._scaler = scaler
        self._xlabel = xlabel
        self._ylabel = ylabel
        self._label = f'S{peak["type"]}'
        
    def _change_units(self):
        for dim in self._vdims:
            dim.label = dim.label.replace('[ns]', '[µs]')
        
    def plot_peak(self,
                  peak,
                  opts_area=immutabledict(),
                  opts_curve=immutabledict(),
                  label='Peak',
                  time_in_us=False,
                  amplitude_prefix='',
                  ):
        """Function which plots a single peak. 
        """
        plot = self.plot_peaks(peak,
                               opts_curve=opts_curve,
                               opts_area=opts_area,
                               time_in_us=time_in_us,
                               label=label,
                               group_label='Peak',
                               amplitude_prefix=amplitude_prefix,
                               _relative_start_time=peak['time'],
                               )
        #FIY: If opts is applied only with an empty dict all previous 
        # options seem to be reseted to default. 
        return plot.opts(**self.opts_dict, legend_limit=10)
    
    
    def plot_peaks(self,
                   peaks,
                   label='Peaks',
                   group_label='Event',
                   opts_area=immutabledict(),
                   opts_curve=immutabledict(),
                   time_in_us=False,
                   time_prefix='',
                   amplitude_prefix='',
                   _relative_start_time=0
                   ):
        if type(peaks) is np.void:
            # In case of supplying only a single peak we have to make 
            # an array again:
            peaks = np.array([peaks])

        if not (np.all(peaks['type'] == peaks[0]['type'])):
            raise ValueError('All peaks must be of the same type (S1, S2 or Unknown)!')

        self.time_in_us = time_in_us
        area, curve = self._init_peak_plot(peaks, amplitude_prefix, time_prefix)
        for ind, peak in enumerate(peaks):
            _area, _curve = self._plot_peak(
                peak,
                label=label + f'_{ind}',
                group_label=group_label,
                opts_curve=opts_curve,
                opts_area=opts_area,
                _relative_start_time=_relative_start_time,
            )
            area *= _area
            curve *= _curve.opts(tools=['vline'])
        return (area * curve).opts(**self.opts_dict, legend_limit=100)
        
        
    def _is_time_based_quantity(self, field_name):
        """Checks if fiel_name belongs to a time based peak property to
        rescale ns value to µs for S2s.
        """
        _is_true = ('time' in field_name 
                    and ((field_name != 'time') 
                         and (field_name != 'endtime')
                         and (field_name != 'center_time')
                        )
                   )
        _is_true |= ('width' in field_name)
        _is_true |= ('range' in field_name)
        return _is_true
        


import strax
import numpy as np

from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.merged_s2s.merged_s2s import MergedS2s

export, __all__ = strax.exporter()


@export
class MergedS2sHighEnergy(MergedS2s):
    __doc__ = HE_PREAMBLE + MergedS2s.__doc__
    depends_on = ('peaklets_he', 'peaklet_classification_he')
    data_kind = 'merged_s2s_he'
    provides = 'merged_s2s_he'
    __version__ = '0.1.0'
    child_plugin = True

    # We cannot, we only have the top array, so should not.
    sum_waveform_top_array = False

    @property
    def n_tpc_pmts(self):
        # Have to hack the url config to avoid nasty numba errors for the main MergedS2s plugin
        return self.n_he_pmts

    def infer_dtype(self):
        return strax.unpack_dtype(self.deps['peaklets_he'].dtype_for('peaklets_he'))

    def compute(self, peaklets_he):
        # There are not any lone hits for the high energy channel,
        #  so create a dummy for the compute method.
        lone_hits = np.zeros(0, dtype=strax.hit_dtype)
        return super().compute(peaklets_he, lone_hits)



from straxen.plugins.events_nv import nVETOEventsSync

import strax

export, __all__ = strax.exporter()


@export
class mVETOEventSync(nVETOEventsSync):
    """
    Plugin which computes synchronized timestamps for the muon-veto with
    respect to the TPC.
    """
    depends_on = ('events_mv', 'detector_time_offsets')
    delay_field_name = 'time_offset_mv'

    provides = 'events_sync_mv'
    __version__ = '0.0.1'
    child_plugin = True

    def compute(self, events_mv, detector_time_offsets):
        return super().compute(events_mv, detector_time_offsets)



from . import hitfinder_thresholds
from .hitfinder_thresholds import *

from . import xenon1t_url_configs
from .xenon1t_url_configs import *

from . import plugins_1t
from .plugins_1t import *

from . import contexts_1t
from .contexts_1t import *



from immutabledict import immutabledict
import warnings
import panel as pn
import numpy as np
import holoviews as hv
import strax
from .holoviews_inspector import EventStats
from .holoviews_peak_data import PlotPeaksTPC
from .holoviews_pmt_array import PlotPMTArrayTPC

class InteractiveTPCEventDisplay():
    """Interactive event display for the TPC. Can be used for waveform
    watching/inspection. For waveform inspection the user can provide
    a subset of peak data which only belongs to the peaks of interest.
    In this case the inspector plot plots all events, but only the 
    selected events with peaks can be choosen with the event slider.
    
    :param events:
    :param peaks:
    :param selection:
    :param gains: Can be used to apply to_pe value to area_per_channel
        before plotting. PMTs with a to_pe value of zero are displayed
        in gray.
    """
    _current_event_index = -1
    
    def __init__(self, 
                 events, 
                 peaks,
                 gains=None,
                 plot_alt_peaks=True,
                ):
        self._events = events
        self._peaks = peaks
        self._plot_alt_peaks = plot_alt_peaks
        self._split_peaks_by_event()
        self._make_event_slider()
        self.gains = gains
        
    def event_display(self, 
                      inspector_config=immutabledict(),
                      exclude_fields=None,
                      inspection_mode=True,
                      plot_alt=True,
                      log_hit_pattern=True,
                      row_height=300,
                      row_width=900,
                     ):
        """Function which plots interactive event display.
        """
        
        plots = self._get_display_componnts(
            inspector_config=inspector_config,
            exclude_fields=exclude_fields,
            inspection_mode=inspection_mode,
            plot_alt=plot_alt,
            log_hit_pattern=log_hit_pattern,
        )             
        pattern, inspector_plot, inspector_widget, s2_peak, s1_peak, event_display = plots
        
        
        event_display.opts(height=row_height, width=row_width)
        row_setting = dict(height=row_height, width=row_width//3)
        s1_s2_pattern = pn.Row(s1_peak.opts(**row_setting), 
                               s2_peak.opts(**row_setting), 
                               pattern.opts(**row_setting))
        if inspection_mode:
            inspector_widget.width = row_width//2
            inspector_widget.height = row_height
            inspector_row = pn.Row(inspector_plot.opts(height=row_height, 
                                                       width=row_width//2),
                                   pn.Column(self.event_slider, inspector_widget)
                                  )
            return pn.Column(s1_s2_pattern, event_display, inspector_row)
        return pn.Column(self.event_slider, s1_s2_pattern, event_display)
        
    
    def _get_display_componnts(self, 
                               inspector_config=immutabledict(),
                               exclude_fields=None,
                               inspection_mode=True,
                               plot_alt=True,
                               log_hit_pattern=True,
                              ):
        """Function which creates all required plot and widget 
        components for the interactive event display. 
        """
        inspector_plot = None
        inspector_widget = None
        if inspection_mode:
            inspector_plot, inspector_widget = self._get_inspector_plot_and_widget(
                inspector_config, 
                exclude_fields
            )
        
        streams = {'event_index': self.event_slider}
        pattern = hv.DynamicMap(self._pmt_pattern_callback, 
                                streams=streams)
        
        s2_peak = hv.DynamicMap(
            self._s2_callback,
            streams=streams)
        s1_peak = hv.DynamicMap(
            self._s1_callback, 
            streams=streams)
        event_display = hv.DynamicMap(
            self._event_callback,
            streams=streams)
        
        return pattern, inspector_plot, inspector_widget, s2_peak, s1_peak, event_display
            
    def _get_inspector_plot_and_widget(self, inspector_config, exclude_fields):
        inspector = EventStats(self._events, 
                               inspector_config,
                               exclude_fields,
                               _index_slider=self.event_slider,
                              )
        stream_hist, stream_marker = inspector._create_streams()
        hist = hv.DynamicMap(inspector._plot_hist_data, 
                             streams=stream_hist)
        selected_points = hv.DynamicMap(inspector._plot_selected_event, 
                                        streams=stream_marker)

        inspector_plot = (hist*selected_points)
        inspector_widget = inspector._make_widget_panel(include_index_slider=False)
        return inspector_plot, inspector_widget
    
    def _selcet_current_event(self, event_index):
        """Function which selects current event. If event has been 
        already selected do nothing. This improves performance as we
        need four different callbacks.
        """
        if self._current_event_index == event_index:
            return
        
        self._selected_event = self._events[event_index]
        self._peaks_in_event = self._peaks[self._fully_contained_index == event_index]
        
        self._s2 = self._peaks_in_event[self._selected_event['s2_index']]
        self._s1 = self._peaks_in_event[self._selected_event['s1_index']]
        other_peaks_index = np.arange(len(self._peaks_in_event))
        mask_other_s2 = self._peaks_in_event['type'] == 2
        mask_other_s2 &= (other_peaks_index != self._selected_event['s2_index'])
        
        self._alt_s2 = None
        _can_plot_alt_s2 = (self._selected_event['alt_s2_index'] != -1) and self._plot_alt_peaks
        if _can_plot_alt_s2:
            self._alt_s2 = self._peaks_in_event[self._selected_event['alt_s2_index']]
            mask_other_s2 &= (other_peaks_index != self._selected_event['alt_s2_index'])
        
        self._alt_s1 = None
        _can_plot_alt_s1 = (self._selected_event['alt_s1_index'] != -1) and self._plot_alt_peaks
        if _can_plot_alt_s1:    
            self._alt_s1 = self._peaks_in_event[self._selected_event['alt_s1_index']]
        
        self._other_s2 = self._peaks_in_event[mask_other_s2]
        self._current_event_index = event_index
        
    
    def _pmt_pattern_callback(self, event_index, top_pmt_array=True, pattern_ops=immutabledict()):
        """Callback function for the PMT hitpattern. Selects main and 
        alt S2 position for a given event_index and creates the plot.
        """
        self._selcet_current_event(event_index)
        
        pmt_array = PlotPMTArrayTPC(
            gains=self.gains,
            top_pmt_array=top_pmt_array,
        )
        
        s2_plot = pmt_array.plot_pmt_array(self._s2, 
                                           label='S2', 
                                           logz=True, #TODO make me optional?
                                           **pattern_ops,
                                          )
        s2_point = hv.Points(
            (self._selected_event['x'], 
             self._selected_event['y']), 
            label='S2').opts(color='red', size=5)
        position_plot = s2_plot
        position_plot = (s2_plot * s2_point)
        
        if not self._alt_s2 is None:
            alt_s2_plot = pmt_array.plot_pmt_array(self._alt_s2, 
                                                   label='alt. S2', 
                                                   logz=True, #TODO make me optional?
                                                   **pattern_ops,
                                                  )
            alt_s2_point = hv.Points(
                (self._selected_event['alt_s2_x'], 
                 self._selected_event['alt_s2_y']), 
                label='alt. S2').opts(color='orange', size=5,) #Todo allow to cutsomize better...
            # Cannot use *= because changes plot order
            position_plot = (alt_s2_plot * alt_s2_point) * position_plot
        
        return position_plot.opts(legend_opts={"click_policy": "hide"})
   
    
    def _peak_callback(self, 
                       main_peak,
                       main_label,
                       alt_peak, 
                       alt_label,
                       time_in_µs=False,
                       opts_curve=immutabledict(),
                       opts_area=immutabledict(),
                       amplitude_prefix=''
                      ):
        """Call back which plots main/alt S1/S2 for display.
        """
        peak_ploter = PlotPeaksTPC()
        peak_plot = peak_ploter.plot_peak(main_peak,
                                          label=main_label,
                                          opts_curve=opts_curve,
                                          opts_area=opts_area,
                                          amplitude_prefix=amplitude_prefix,
                                          time_in_us=time_in_µs)
        if not alt_peak is None:
            opts_curve = self._add_legend_mute(opts_curve)
            opts_area = self._add_legend_mute(opts_area)
            alt_peak_plot = peak_ploter.plot_peak(alt_peak,
                                                  label=alt_label,
                                                  opts_curve=opts_curve,
                                                  opts_area=opts_area,
                                                  amplitude_prefix=amplitude_prefix,
                                                  time_in_us=time_in_µs)
        
            peak_plot *= alt_peak_plot
        return peak_plot
        
    
    def _s2_callback(self, 
                     event_index,
                     opts_peak_area=immutabledict(color='purple'),
                     opts_peak_curve=immutabledict(color='purple'),
                    ):
        """Call back which plots main/alt S2 for display.
        """
        self._selcet_current_event(event_index)
        return self._peak_callback(self._s2, 'S2', 
                                   self._alt_s2, 'alt. S2', 
                                   time_in_µs=True,
                                   opts_area=opts_peak_area,
                                   opts_curve=opts_peak_curve,
                                   amplitude_prefix='S2',
                                  ).opts(title='main/alt. S2')
    
    def _s1_callback(self, 
                     event_index,
                     opts_peak_area=immutabledict(color='orange'),
                     opts_peak_curve=immutabledict(color='orange'),
                    ):
        """Call back which plots main/alt S1 for display.
        """
        self._selcet_current_event(event_index)
        return self._peak_callback(self._s1, 'S1', 
                                   self._alt_s1, 'alt. S1', 
                                   time_in_µs=False,
                                   opts_area=opts_peak_area,
                                   opts_curve=opts_peak_curve,
                                   amplitude_prefix='S1',
                                  ).opts(title='main/alt. S1')
    
    def _event_callback(self, 
                        event_index,
                       ):
        """Callback function to plot the main display.
        """
        self._selcet_current_event(event_index)
        
        # Get event start time and extend a bit to the left:
        event_start = self._selected_event['time']
        event_end = self._selected_event['endtime']
        s1_start = self._selected_event['s1_time']
        rel_start = max(event_start, s1_start)
        peak_ploter = PlotPeaksTPC()                
        event_plot = []
        
        mask = self._peaks_in_event['type'] == 2
        self.n_s2 = np.sum(mask)
        if np.any(mask):
            plot_s2 = peak_ploter.plot_peaks(self._peaks_in_event[mask],
                                             label=f'S2 {event_index}',
                                             group_label=f'Event {event_index}',
                                             opts_curve={'color': 'purple', 'muted_alpha': 0,},
                                             opts_area={'color': 'purple', 'muted_alpha': 0,},
                                             time_in_us=True,
                                             time_prefix='Event',
                                             amplitude_prefix='Event',
                                             _relative_start_time=rel_start)
            event_plot.append(plot_s2)

        mask = self._peaks_in_event['type'] == 1
        self.n_s1 = np.sum(mask)
        if np.any(mask):
            plot_s1 = peak_ploter.plot_peaks(self._peaks_in_event[mask],
                                             label=f'S1 {event_index}',
                                             group_label=f'Event {event_index}',
                                             opts_curve={'color': 'orange', 'muted_alpha': 0,},
                                             opts_area={'color': 'orange', 'muted_alpha': 0,},
                                             time_in_us=True,
                                             time_prefix='Event',
                                             amplitude_prefix='Event',
                                             _relative_start_time=rel_start)
            event_plot.append(plot_s1)


        mask = self._peaks_in_event['type'] == 0
        self.n_s0 = np.sum(mask)
        if np.any(mask):
            plot_s0 = peak_ploter.plot_peaks(self._peaks_in_event[mask],
                                             label=f'S0 {event_index}',
                                             group_label=f'Event {event_index}',
                                             opts_curve={'color': 'gray',  'muted_alpha': 0,},
                                             opts_area={'color': 'gray', 'muted_alpha': 0,},
                                             time_in_us=True,
                                             time_prefix='Event',
                                             amplitude_prefix='Event',
                                             _relative_start_time=rel_start)
            event_plot.append(plot_s0)
        
        shaded_regions = self._get_alt_main_shading(self._selected_event, 
                                                    rel_start, in_µs=True)
        event_plot = hv.Overlay(event_plot + shaded_regions)
        return event_plot.opts(legend_opts={"click_policy": "hide",}, 
                               show_legend=True,
                               xlim=((event_start-rel_start)/10**3, 
                                     (event_end-rel_start)/10**3),
                               ylim=(0, None),
                              )
    
    
    def _get_alt_main_shading(self, event, start_time, in_µs=True):
        """Function which shades alt./main S1/S2 with 
        """
        time_scaler = 1
        if in_µs:
            time_scaler = 1000

        peak_times = dict()
        for peak in ['s1', 's2', 'alt_s1', 'alt_s2']:
            peak_start = event['_'.join((peak, 'time'))]
            peak_end = event['_'.join((peak, 'endtime'))]
            _has_peak = peak_start != -1
            if _has_peak:
                peak_times[peak] = ((peak_start - start_time)/time_scaler,
                                    (peak_end - start_time)/time_scaler,
                                    )
            else:
                peak_times[peak] = (-start_time, -start_time)

        shaded_regions = []
        for peak_type, (start, end) in peak_times.items():
            span = hv.VSpan(start, end).opts(color='gray', alpha=0.4)
            text = hv.Text(
                start, 0, rotation=90,
                text=peak_type, valign='bottom', halign='left').opts(text_alpha=1)

            # In case peak does not exists make label transparent. (We have 
            # to always make 4 labels as otherwise hv.Text breaks if we 
            # start with less peaks.)
            _no_peak = start == -start_time
            if _no_peak:
                text.opts(text_alpha=0)

            shaded_regions += [span, text]  

        return shaded_regions
        
    
    @staticmethod
    def _add_legend_mute(options):
        options = {k: v for k, v in options.items()}
        options['muted'] = True
        options['muted_alpha'] = 0
        return immutabledict(options)
        
    def _split_peaks_by_event(self):
        """Split provided peaks by event.
        """
        fci = strax.fully_contained_in(self._peaks, self._events)
        if np.any(fci == -1):
            warnings.warn('Found peaks which are not belonging to any '
                          'event. We are dropping these peaks for efficiency '
                          'purposes.')
        self._peaks = self._peaks[fci != -1]
        self._fully_contained_index = fci[fci != -1]
        self._events_with_peaks = list(np.unique(self._fully_contained_index))
    
    
    def _make_event_slider(self):
        """Creates slider for all events with peak data.
        """
        self.event_slider = pn.widgets.DiscreteSlider(
            name='Selected Event Index:',
            options=self._events_with_peaks, 
            value=self._events_with_peaks[0],
            value_throttled=True,
        )



import strax
from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.peaks.peak_basics import PeakBasics


export, __all__ = strax.exporter()


@export
class PeakBasicsHighEnergy(PeakBasics):
    __doc__ = HE_PREAMBLE + PeakBasics.__doc__
    __version__ = '0.0.2'
    depends_on = 'peaks_he'
    provides = 'peak_basics_he'
    child_ends_with = '_he'

    def compute(self, peaks_he):
        return super().compute(peaks_he)



import strax
import straxen
from straxen.plugins.peaks._peak_s1_positions_base import PeakS1PositionBase


export, __all__ = strax.exporter()


@export
class PeakS1PositionCNN(PeakS1PositionBase):
    """
    S1 CNN for (x,y,z) position S1 reconstruction at peak level
    """
    provides = "peak_s1_positions_cnn"
    algorithm = "s1_cnn"
    __version__ = '0.0.1'

    tf_model_s1_cnn = straxen.URLConfig(
        default=f'tf://'
                f'resource://'
                f'xedocs://posrec_models'
                f'?version=ONLINE'
                f'&run_id=plugin.run_id'
                f'&kind=s1_cnn'
                f'&fmt=abs_path'
                f'&attr=value',
        help='s1 position 3d reconstruction cnn model. Should be opened using the "tf" descriptor. '
             'Set to "None" to skip computation',
        cache=3,
    )



import os
import json
import tempfile
import pandas as pd
import strax
import straxen
import fsspec
import utilix.rundb
from straxen.test_utils import nt_test_context, nt_test_run_id
import unittest
import pickle
import random
import numpy as np
from datetime import datetime


class DummyObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


@straxen.URLConfig.register('random')
def generate_random(_):
    return random.random()


@straxen.URLConfig.register('unpicklable')
def return_lamba(_):
    return lambda x: x


@straxen.URLConfig.register('large-array')
def large_array(_):
    return np.ones(1_000_000).tolist()


@straxen.URLConfig.register('object-list')
def object_list(length):
    length = int(length)
    return [DummyObject(a=i, b=i+1) for i in range(length)]


@straxen.URLConfig.preprocessor
def formatter(config, **kwargs):
    if not isinstance(config, str):
        return config
    try:
        config = config.format(**kwargs)
    except KeyError:
        pass
    return config


GLOBAL_VERSIONS = {
    'global_v1': {
        'test_config': 'v0'
    }
}


@straxen.URLConfig.preprocessor
def replace_global_version(config, name=None, **kwargs):
    if name is None:
        return

    if not isinstance(config, str):
        return config

    if not straxen.URLConfig.SCHEME_SEP in config:
        return config

    version = straxen.URLConfig.kwarg_from_url(config, 'version')

    if version is None:
        return config

    if version.startswith('global_') and version in GLOBAL_VERSIONS:
        version = GLOBAL_VERSIONS[version].get(name, version)
        config = straxen.URLConfig.format_url_kwargs(config, version=version)
    return config


class ExamplePlugin(strax.Plugin):
    depends_on = ()
    dtype = strax.time_fields
    provides = ('test_data',)
    test_config = straxen.URLConfig(default=42,)
    cached_config = straxen.URLConfig(default=666, cache=1)

    def compute(self):
        pass


class TestURLConfig(unittest.TestCase):
    def setUp(self):
        st = nt_test_context()
        st.register(ExamplePlugin)
        self.st = st

    def test_default(self):
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 42)

    def test_literal(self):
        self.st.set_config({'test_config': 666})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 666)

    def test_leading_zero_int(self):
        self.st.set_config({'test_config': 'format://{value}?value=0666'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, '0666')
        self.assertIsInstance(p.test_config, str)


    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_cmt_protocol(self):
        self.st.set_config({'test_config': 'cmt://elife?version=v1&run_id=plugin.run_id'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertTrue(abs(p.test_config-219203.49884000001) < 1e-2)

    def test_json_protocol(self):
        self.st.set_config({'test_config': 'json://[1,2,3]'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, [1, 2, 3])

    def test_format_protocol(self):
        self.st.set_config({'test_config': 'format://{run_id}?run_id=plugin.run_id'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, nt_test_run_id)

    def test_fsspec_protocol(self):
        with fsspec.open('memory://test_file.json', mode='w') as f:
            json.dump({"value": 999}, f)
        self.st.set_config(
            {'test_config': 'take://json://fsspec://memory://test_file.json?take=value'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 999)

    def test_chained(self):
        self.st.set_config({'test_config': 'take://json://[1,2,3]?take=0'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 1)

    def test_take_nested(self):
        self.st.set_config({'test_config': 'take://json://{"a":[1,2,3]}?take=a&take=0'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 1)

    @unittest.skipIf(not straxen.utilix_is_configured(),
                     "No db access, cannot test!")
    def test_bodedga_get(self):
        """Just a didactic example"""
        self.st.set_config({
            'test_config':
                'take://'
                'resource://'
                'XENONnT_numbers.json'
                '?fmt=json'
                '&take=g1'
                '&take=v2'
                '&take=value'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        # Either g1 is 0, bodega changed or someone broke URLConfigs
        self.assertTrue(p.test_config)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_itp_dict(self, ab_value=20, cd_value=21, dump_as='json'):
        """
        Test that we are getting ~the same value from interpolating at the central date in a dict

        :param ab_value, cd_value: some values to test against
        :param dump_as: Write as csv or as json file
        """
        central_datetime = utilix.rundb.xent_collection().find_one(
            {'number': int(nt_test_run_id)},
            projection={'start': 1}
        ).get('start', 'QUERY FAILED!')
        fake_file = {'time': [datetime(2000, 1, 1).timestamp() * 1e9,
                              central_datetime.timestamp() * 1e9,
                              datetime(2040, 1, 1).timestamp() * 1e9],
                     'ab': [10, ab_value, 30],
                     'cd': [11, cd_value, 31]
                         }

        temp_dir = tempfile.TemporaryDirectory()

        if dump_as == 'json':
            fake_file_name = os.path.join(temp_dir.name, 'test_seg.json')
            with open(fake_file_name, 'w') as f:
                json.dump(fake_file, f)
        elif dump_as == 'csv':
            # This example also works well with dataframes!
            fake_file_name = os.path.join(temp_dir.name, 'test_seg.csv')
            pd.DataFrame(fake_file).to_csv(fake_file_name)
        else:
            raise ValueError

        self.st.set_config({'test_config':f'itp_dict://'
                                          f'resource://'
                                          f'{fake_file_name}'
                                          f'?run_id=plugin.run_id'
                                          f'&fmt={dump_as}'
                                          f'&itp_keys=ab,cd'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertIsInstance(p.test_config, dict)
        assert np.isclose(p.test_config['ab'], ab_value, rtol=1e-3)
        assert np.isclose(p.test_config['cd'], cd_value, rtol=1e-3)
        temp_dir.cleanup()

    def test_itp_dict_csv(self):
        self.test_itp_dict(dump_as='csv')

    def test_rekey(self):
        original_dict = {'a': 1, 'b': 2, 'c': 3}
        check_dict = {'anew': 1, 'bnew': 2, 'cnew': 3}

        temp_dir = tempfile.TemporaryDirectory()

        fake_file_name = os.path.join(temp_dir.name, 'test_dict.json')
        with open(fake_file_name, 'w') as f:
            json.dump(original_dict, f)

        self.st.set_config({'test_config': f'rekey_dict://resource://{fake_file_name}?'
                                           f'fmt=json&replace_keys=a,b,c'
                                           f'&with_keys=anew,bnew,cnew'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, check_dict)
        temp_dir.cleanup()

    def test_print_protocol_desc(self):
        straxen.URLConfig.print_protocols()

    def test_cache(self):
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')

        # sanity check that default value is not affected
        self.assertEqual(p.cached_config, 666)
        self.st.set_config({'cached_config': 'random://abc'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')

        # value is randomly generated when accessed so if
        # its equal when we access it again, its coming from the cache
        cached_value = p.cached_config
        self.assertEqual(cached_value, p.cached_config)

        # now change the config to which will generate a new number
        self.st.set_config({'cached_config': 'random://dfg'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')

        # sanity check that the new value is still consistent i.e. cached
        self.assertEqual(p.cached_config, p.cached_config)

        # test if previous value is evicted, since cache size is 1
        self.assertNotEqual(cached_value, p.cached_config)

        # verify pickalibility of objects in cache dont affect plugin pickalibility
        self.st.set_config({'cached_config': 'unpicklable://dfg'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        with self.assertRaises(AttributeError):
            pickle.dumps(p.cached_config)
        pickle.dumps(p)

    def test_cache_size(self):
        '''test the cache helper functions
        '''
        # make sure the value has a detectable size
        self.st.set_config({'cached_config': 'large-array://dfg'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')

        # fetch the value so its stored in the cache
        value = p.cached_config

        # cache should now have finite size
        self.assertGreater(straxen.config_cache_size_mb(), 0.0)

        # test if clearing cache works as expected
        straxen.clear_config_caches()
        self.assertEqual(straxen.config_cache_size_mb(), 0.0)

    def test_filter_kwargs(self):
        all_kwargs = dict(a=1, b=2, c=3)

        # test a function that takes only a seubset of the kwargs
        def func1(a=None, b=None):
            return
        
        filtered1 = straxen.filter_kwargs(func1, all_kwargs)
        self.assertEqual(filtered1, dict(a=1, b=2))
        func1(**filtered1)


        # test function that accepts wildcard kwargs
        def func2(**kwargs):
            return
        filtered2 = straxen.filter_kwargs(func2, all_kwargs)
        self.assertEqual(filtered2, all_kwargs)
        func2(**filtered2)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_dry_evaluation(self):
        """
        Check that running a dry evaluation can be done outside of the
        context of a URL config and yield the same result.
        """
        plugin_url = 'cmt://electron_drift_velocity?run_id=plugin.run_id&version=v3'
        self.st.set_config({'test_config': plugin_url})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        correct_val = p.test_config

        # We can also get it from one of these methods
        dry_val1 = straxen.URLConfig.evaluate_dry(
            f'cmt://electron_drift_velocity?run_id={nt_test_run_id}&version=v3')
        dry_val2 = straxen.URLConfig.evaluate_dry(
            f'cmt://electron_drift_velocity?version=v3', run_id=nt_test_run_id)

        # All methods should yield the same
        assert correct_val == dry_val1 == dry_val2

        # However dry-evaluation does NOT allow loading the plugin.run_id
        # as in the plugin_url and should complain about that
        with self.assertRaises(ValueError):
            straxen.URLConfig.evaluate_dry(plugin_url)

    def test_objects_to_dict(self):
        n = 3
        self.st.set_config({'test_config': f'objects-to-dict://object-list://{n}?key_attr=a&value_attr=b'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, {i:i+1 for i in range(n)})

    def test_list_to_array(self):
        n = 3
        self.st.set_config({'test_config': f'list-to-array://object-list://{n}'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertIsInstance(p.test_config, np.ndarray)

    def test_format_preprocessor(self):
        self.st.set_config({'test_config': '{name}:{run_id}'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, f'test_config:{nt_test_run_id}')
        self.assertEqual(p.test_config, p.config['test_config'])

    def test_global_version_preprocessor(self):
        self.st.set_config({'test_config': 'fake://url?version=global_v1'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 'fake://url?version=v0')

    def test_alphabetize_url_kwargs(self):
        """URLConfig preprocessor to rearange the order of arguments given buy a 
        url to ensure the same url with a different hash order gives the same hash"""
        url = "xedocs://electron_lifetimes?run_id=034678&version=v5&attr=value"
        intended_url = "xedocs://electron_lifetimes?attr=value&run_id=034678&version=v5"
        preprocessed_url = straxen.url_config.alphabetize_url_kwargs(url)
        self.assertEqual(intended_url, preprocessed_url)

    def test_xedocs_global_version_hash_coinsistency(self):
        # Same URLs but the queries are in a different order
        st1 = self.st.new_context(config={"test_config":"fake://electron_lifetimes?run_id=25000&version=v5&attr=value"})
        st2 = self.st.new_context(config={"test_config":"fake://electron_lifetimes?attr=value&run_id=25000&version=v5"})
        self.assertEqual(st1.key_for(25000, 'corrected_areas').lineage_hash, st2.key_for(25000, 'corrected_areas').lineage_hash)
    
    def test_global_version_not_changed(self):
        """
          - if no global version is matched, the url version should not be changed
          - if config is not matched, the url version should not be changed
        """
        assert 'global_v2' not in GLOBAL_VERSIONS
        self.st.set_config({'test_config': 'fake://url?version=global_v2'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 'fake://url?version=global_v2')
           
        class ExamplePluginNew(ExamplePlugin):
            test_config_new = straxen.URLConfig(default='rootisthesourceofallevil',)

        self.st.register(ExamplePluginNew)
        self.st.set_config({'test_config_new': 'fake://url?version=global_v1'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config_new, 'fake://url?version=global_v1')

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test run_doc protocol.")
    def test_run_doc_protocol(self):
        self.st.set_config({ 'test_config': "run_doc://mode?run_id=plugin.run_id" })
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 'tpc_commissioning')

        self.st.set_config({ 'test_config': "run_doc://fake_key?run_id=plugin.run_id&default=42" })
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(p.test_config, 42)

        with self.assertRaises(ValueError):
            self.st.set_config({ 'test_config': "run_doc://mode?run_id=plugin.run_id" })
            p = self.st.get_single_plugin(999999999, 'test_data')
            return p.test_config

    def test_pad_array(self):
        """Test that pad_array works as expected
        """

        self.st.set_config({'test_config': 
                            'pad-array://json://[1,2,3]?pad_left=2&pad_right=3&pad_value=0'})
        p = self.st.get_single_plugin(nt_test_run_id, 'test_data')
        self.assertEqual(len(p.test_config), 8)
        self.assertEqual(p.test_config[0], 0)
        self.assertEqual(p.test_config[-1], 0)



import numpy as np
from hypothesis import given, settings
import hypothesis.strategies as strat
import strax
from strax.testutils import fake_hits
import straxen
from straxen.plugins.peaklets.peaklets import get_tight_coin


@settings(deadline=None)
@given(strat.lists(strat.integers(min_value=0, max_value=10),
                   min_size=8, max_size=8, unique=True),
       )
def test_create_outside_peaks_region(time):
    time = np.sort(time)
    time_intervals = np.zeros(len(time)//2, strax.time_dt_fields)
    time_intervals['time'] = time[::2]
    time_intervals['length'] = time[1::2] - time[::2]
    time_intervals['dt'] = 1

    st = straxen.contexts.demo()
    p = st.get_single_plugin('0', 'peaklets')
    outside = p.create_outside_peaks_region(time_intervals, 0, np.max(time))

    touching = strax.touching_windows(outside, time_intervals, window=0)

    for tw in touching:
        print(tw)
        assert np.diff(tw) == 0, 'Intervals overlap although they should not!'


def test_n_hits():
    if not straxen.utilix_is_configured():
        return
    records = np.zeros(2, dtype=strax.record_dtype())
    records['length'] = 5
    records['pulse_length'] = 5
    records['dt'] = 1
    records['channel'] = [0, 1]
    records['data'][0, :5] = [0, 1, 1, 0, 1]
    records['data'][1, :5] = [0, 1, 0, 0, 0]

    st = straxen.contexts.xenonnt_online()
    st.set_config({'hit_min_amplitude': 1})
    p = st.get_single_plugin('0', 'peaklets')
    res = p.compute(records, 0, 999)
    peaklets = res['peaklets']
    assert peaklets['n_hits'] == 3, f"Peaklet has the wrong number of hits!"


@given(fake_hits,
       strat.lists(elements=strat.integers(0, 9), min_size=20))
@settings(deadline=None)
def test_get_tight_coin(hits, channel):
    hits['area'] = 1
    hits['channel'] = channel[:len(hits)]  # In case there are less channel then hits (unlikely)
    gap_threshold = 10
    peaks = strax.find_peaks(hits,
                             adc_to_pe=np.ones(10),
                             right_extension=0, left_extension=0,
                             gap_threshold=gap_threshold,
                             min_channels=1,
                             min_area=0)

    peaks_max_time = peaks['time'] + peaks['length']//2
    hits_max_time = hits['time'] + hits['length']//2

    left = 5
    right = 5
    tight_coin_channel = get_tight_coin(hits_max_time,
                                        hits['channel'],
                                        peaks_max_time,
                                        left,
                                        right)
    for ind, p_max_t in enumerate(peaks_max_time):
        m_hits_in_peak = (hits_max_time >= (p_max_t - left))
        m_hits_in_peak &= (hits_max_time <= (p_max_t + right))
        n_channel = len(np.unique(hits[m_hits_in_peak]['channel']))
        assert n_channel == tight_coin_channel[ind], f'Wrong number of tight channel got {tight_coin_channel[ind]}, but expectd {n_channel}'  # noqa



import numpy as np
import straxen
import strax

export, __all__ = strax.exporter()


@export
class S2ReconPosDiff(strax.Plugin):
    """
    Plugin that provides position reconstruction difference for S2s in events, see note:
    https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:shengchao:sr0:reconstruction_quality
    """
    __version__ = '0.0.3'

    parallel = True
    depends_on = 'event_basics'
    provides = 's2_recon_pos_diff'
    save_when = strax.SaveWhen.EXPLICIT

    recon_alg_included = straxen.URLConfig(
        help='The list of all reconstruction algorithm considered.',
        default=('_mlp', '_gcn', '_cnn'), infer_type=False,
    )

    def infer_dtype(self):
        dtype = [
            ('s2_recon_avg_x', np.float32,
             'Mean value of x for main S2'),
            ('alt_s2_recon_avg_x', np.float32,
             'Mean value of x for alternatice S2'),
            ('s2_recon_avg_y', np.float32,
             'Mean value of y for main S2'),
            ('alt_s2_recon_avg_y', np.float32,
             'Mean value of y for alternatice S2'),
            ('s2_recon_pos_diff', np.float32,
             'Reconstructed position difference for main S2'),
            ('alt_s2_recon_pos_diff', np.float32,
             'Reconstructed position difference for alternative S2'),
        ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events):

        result = np.zeros(len(events), dtype=self.dtype)
        result['time'] = events['time']
        result['endtime'] = strax.endtime(events)
        # Computing position difference
        self.compute_pos_diff(events, result)
        return result

    def cal_avg_and_std(self, values, axis=1):
        average = np.mean(values, axis=axis)
        std = np.std(values, axis=axis)
        return average, std

    def eval_recon(self, data, name_x_list, name_y_list):
        """
        This function reads the name list based on s2/alt_s2 and all recon algorithm registered
        Each row consists the reconstructed x/y and their average and standard deviation is calculated
        """
        # lazy fix to delete field name in array, otherwise np.mean will complain
        x_avg, x_std = self.cal_avg_and_std(np.array(data[name_x_list].tolist()))
        y_avg, y_std = self.cal_avg_and_std(np.array(data[name_y_list].tolist()))
        r_std = np.sqrt(x_std ** 2 + y_std ** 2)
        res = x_avg, y_avg, r_std
        return res

    def compute_pos_diff(self, events, result):

        alg_list = self.recon_alg_included
        for peak_type in ['s2', 'alt_s2']:
            # Selecting S2s for pos diff
            # - must exist (index != -1)
            # - must have positive AFT
            # - must contain all alg info
            cur_s2_bool = (events[peak_type + '_index'] != - 1)
            cur_s2_bool &= (events[peak_type + '_area_fraction_top'] > 0)
            for name in self.recon_alg_included:
                cur_s2_bool &= ~np.isnan(events[peak_type + '_x' + name])
                cur_s2_bool &= ~np.isnan(events[peak_type + '_y' + name])

            # default value is nan, it will be overwrite if the event satisfy the requirements
            result[peak_type + '_recon_pos_diff'][:] = np.nan
            result[peak_type + '_recon_avg_x'][:] = np.nan
            result[peak_type + '_recon_avg_y'][:] = np.nan

            if np.any(cur_s2_bool):
                name_x_list = []
                name_y_list = []
                for alg in alg_list:
                    name_x_list.append(peak_type + '_x' + alg)
                    name_y_list.append(peak_type + '_y' + alg)

                # Calculating average x,y, and position difference
                x_avg, y_avg, r_std = self.eval_recon(events[cur_s2_bool], name_x_list, name_y_list)
                result[peak_type + '_recon_pos_diff'][cur_s2_bool] = r_std
                result[peak_type + '_recon_avg_x'][cur_s2_bool] = x_avg
                result[peak_type + '_recon_avg_y'][cur_s2_bool] = y_avg



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakProximity(strax.OverlapWindowPlugin):
    """
    Look for peaks around a peak to determine how many peaks are in
    proximity (in time) of a peak.
    """
    __version__ = '0.4.0'

    depends_on = ('peak_basics',)
    dtype = [
                ('n_competing', np.int32,
                 'Number of nearby larger or slightly smaller peaks'),
                ('n_competing_left', np.int32,
                 'Number of larger or slightly smaller peaks left of the main peak'),
                ('t_to_prev_peak', np.int64,
                 'Time between end of previous peak and start of this peak [ns]'),
                ('t_to_next_peak', np.int64,
                 'Time between end of this peak and start of next peak [ns]'),
                ('t_to_nearest_peak', np.int64,
                 'Smaller of t_to_prev_peak and t_to_next_peak [ns]')
            ] + strax.time_fields

    min_area_fraction = straxen.URLConfig(
        default=0.5, infer_type=False,
        help='The area of competing peaks must be at least '
             'this fraction of that of the considered peak')

    nearby_window = straxen.URLConfig(
        default=int(1e7), infer_type=False,
        help='Peaks starting within this time window (on either side)'
             'in ns count as nearby.')

    peak_max_proximity_time = straxen.URLConfig(
        default=int(1e8), infer_type=False,
        help='Maximum value for proximity values such as t_to_next_peak [ns]')

    def get_window_size(self):
        return self.peak_max_proximity_time

    def compute(self, peaks):
        windows = strax.touching_windows(peaks, peaks,
                                         window=self.nearby_window)
        n_left, n_tot = self.find_n_competing(
            peaks,
            windows,
            fraction=self.min_area_fraction)

        t_to_prev_peak = (
                np.ones(len(peaks), dtype=np.int64)
                * self.peak_max_proximity_time)
        t_to_prev_peak[1:] = peaks['time'][1:] - peaks['endtime'][:-1]

        t_to_next_peak = t_to_prev_peak.copy()
        t_to_next_peak[:-1] = peaks['time'][1:] - peaks['endtime'][:-1]

        return dict(
            time=peaks['time'],
            endtime=strax.endtime(peaks),
            n_competing=n_tot,
            n_competing_left=n_left,
            t_to_prev_peak=t_to_prev_peak,
            t_to_next_peak=t_to_next_peak,
            t_to_nearest_peak=np.minimum(t_to_prev_peak, t_to_next_peak))

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def find_n_competing(peaks, windows, fraction):
        n_left = np.zeros(len(peaks), dtype=np.int32)
        n_tot = n_left.copy()
        areas = peaks['area']

        for i, peak in enumerate(peaks):
            left_i, right_i = windows[i]
            threshold = areas[i] * fraction
            n_left[i] = np.sum(areas[left_i:i] > threshold)
            n_tot[i] = n_left[i] + np.sum(areas[i + 1:right_i] > threshold)

        return n_left, n_tot



import numpy as np
import numba
import strax
import straxen
from strax.processing.peak_splitting import natural_breaks_gof

export, __all__ = strax.exporter()


@export
class LocalMinimumInfo(strax.LoopPlugin):
    """
    Looks for the main S2 peak in an event, finds the local minimum (if there is one),
    and looks to compute some figures of merit such as the max goodness of split,
    width of the valley, and height of the valley.
    """

    depends_on = ('event_basics', 'peaks')
    data_kind = 'events'
    provides = 'event_local_min_info'
    parallel = 'process'
    compressor = 'zstd'

    __version__ = '0.1.1'
    dtype = strax.time_fields + [(('Maximum Goodness of Split',
                                   's2_max_gos'), np.float32),
                                 (('Number of local maxima of the smoothed peak',
                                   's2_num_loc_max'), np.int16),
                                 (('Full gap at p% of the valley height of the deepest valley [ns],'
                                   'by default p = 90', 's2_valley_gap'), np.float32),
                                 (('Valley depth over max height of the deepest valley',
                                   's2_valley_height_ratio'), np.float32)]

    divide_90p_width_localmin = straxen.URLConfig(
        default=7., type=float,
        help="The peak is smoothed by dividing the 90p width by"
             "this number, and coverting it into number of samples."
             "This is then the 'n' used in the smoothing kernel"
             "shown below."
    )

    smoothing_power_localmin = straxen.URLConfig(
        default=3., type=float,
        help="The power used in the smoothing filter with a kernel of"
             "(1-(x/n)^p)^p, where p is the power"
    )

    percentage_threshold_localmin = straxen.URLConfig(
        default=0.1, type=float,
        help="The height threshold for the peak as a percentage"
             "of the maximum, used to reject the low parts"
             "of the peak in order to find the local extrema."
    )

    percent_valley_height = straxen.URLConfig(
        default=0.9, type=float,
        help="The percentage of the valley height of the deepest"
             "valley for which to calculate the valley width"
    )

    def compute_loop(self, event, peaks):
        """
        This finds the maxima and minima for the main S2 peak and calculates its info such as
        the number of local maxima, the depth of the deepest local minimum over the maximum height
        of the peak (s2_valley_height_ratio), and the width of the local minimum valley at
        90% of the valley height.
        :param event: The event
        :param peaks: The peaks belonging to the event, only the main S2 peak is considered,
        if there is none, this plugin returns none
        :return: Returns a dictionary containing all of the fields above for each main S2 peak
        as well as the timing information.
        """
        max_gos = np.nan
        num_loc_maxes = 0
        valley_gap = np.nan
        valley_height_ratio = np.nan

        if event['s2_area'] > 0:
            p = peaks[event['s2_index']]

            smoothing_number = p['width'][9] / p['dt']
            smoothing_number = np.ceil(smoothing_number / self.divide_90p_width_localmin)
            smoothed_peak = power_smooth(p['data'][:p['length']],
                                         int(smoothing_number),
                                         self.smoothing_power_localmin)

            if len(smoothed_peak) > 0:
                # Set data below percentage threshold on both side to zeros
                left, right = bounds_above_percentage_height(smoothed_peak,
                                                             self.percentage_threshold_localmin)

                # Maximum GOS calculation for data above percentage
                max_gos = np.max(natural_breaks_gof(p['data'][left: right],
                                                    p['dt']))

                # Local minimum based information
                maxes, mins = identify_local_extrema(smoothed_peak)

                maxes = maxes[np.logical_and(maxes >= left, maxes < right)]
                mins = mins[np.logical_and(mins >= left, mins < right)]

                num_loc_maxes = len(maxes)

                valley_gap, valley = full_gap_percent_valley(smoothed_peak,
                                                             maxes,
                                                             mins,
                                                             self.percent_valley_height,
                                                             p['dt'])

                valley_height_ratio = valley / np.max(smoothed_peak)

        return {'time': event['time'],
                'endtime': event['endtime'],
                's2_max_gos': max_gos,
                's2_num_loc_max': num_loc_maxes,
                's2_valley_gap': valley_gap,
                's2_valley_height_ratio': valley_height_ratio}


@numba.njit()
def full_gap_percent_valley(smoothp, max_loc, min_loc, pv, dt):
    """
    Full gap at percent valley. The width of the valley at "pv" of the valley height
    Only do this for peaks which have number of maxes-number of mins = 1, since otherwise
    this local minimum finding doesn't make sense. Furthermore, it gets rid of those peaks
    which start at some high value likely because they are the tails of another peak
    or something.
    :param smoothp: The smoothed peak
    :param max_loc: Location of every local maximum of the peak
    :param min_loc: Location of every local minimum of the peak
    :param pv: "Percent value", the percent of the valley height for which to calculate the gap
    :param dt: The time of one sample in ns
    :return: The gap in ns, the depth in PE
    """
    n_gap = len(min_loc)
    p_length = len(smoothp)
    if ~np.logical_and((len(max_loc) - n_gap == 1), (len(min_loc) > 0)):
        return 0, 0
    else:
        gaps = np.zeros(n_gap)
        gap_heights = np.zeros(len(min_loc))
        for j in range(n_gap):
            gh = np.min(smoothp[max_loc[j:j + 2]])
            gh -= smoothp[min_loc[j]]

            height_pv = (smoothp[min_loc[j]] + gh * pv)

            above_hpv = smoothp > height_pv
            above_hpv |= np.arange(p_length) < max_loc[j]
            left_crossing = np.argmin(above_hpv)

            above_hpv = smoothp > height_pv
            above_hpv &= np.arange(p_length) >= min_loc[j]
            right_crossing = np.argmax(above_hpv)

            gaps[j] = right_crossing - left_crossing
            gap_heights[j] = gh

        max_gap = gaps[np.argmax(gap_heights)]
        valley_depth = gap_heights[np.argmax(gap_heights)]
        return max_gap * dt, valley_depth


def bounds_above_percentage_height(p, percent):
    """
    Finding the index bounds of the peak above given percentage
    :param p: The peak
    :param percent: The percentage of the maximum height to cut the peak,
    this is to reject the tails and noise before and after the bulk of the peak
    :return: The left and right (exclusive) index of samples above the percent
    """
    percent_height = np.max(p) * percent
    above_percent_height = np.where(p >= percent_height)[0]

    return above_percent_height[0], above_percent_height[-1] + 1


def identify_local_extrema(smoothp):
    """
    Identifies local minima and maxima by comparing each point to the one before and after it
    :param smoothp: smoothed peak
    :return: The locations of the minima, the locations of the maxima
    """
    larger_than_next = (smoothp > np.pad(smoothp[:-1], (1, 0)))
    larger_than_next = larger_than_next.astype('int')

    max_loc = np.where(np.diff(larger_than_next) < 0)[0]
    min_loc = np.where(np.diff(larger_than_next) > 0)[0]

    return max_loc, min_loc


def power_smooth(origindata, cover_num, power):
    """
    A smoothing filter to get rid of the noise in peaks so that we don't find too many local extrema
    that are just noisy
    :param origindata: Original peak
    :param cover_num: The cover number for smoothing
    high cover numbers mean you smooth over a larger region
    :param power: The power of the smoothing
    essentially tunes how much your smoothing filter looks like a square
    :return: The smoothed waveform
    """
    x_zeroed = np.arange(-cover_num, cover_num + 1)
    weight = (1 - np.abs(x_zeroed / cover_num) ** power) ** power
    weight = weight / np.sum(weight)

    smoothed_data = np.convolve(origindata, weight)[cover_num:-cover_num]
    return smoothed_data



import numpy as np
import strax
import straxen
from warnings import warn
export, __all__ = strax.exporter()

@export
class PeakTopBottomParams(strax.Plugin):
    """
    Pluging that computes timing characteristics of top and bottom waveforms 
    based on waveforms stored at peak level
    """
    depends_on = ('peaks', 'peak_basics')
    provides = 'peak_top_bottom_params'
    __version__ = '0.0.0'

    def infer_dtype(self):
        dtype = []
        peak_basics_fields = self.deps['peak_basics'].dtype.fields
        self.arrs = ['top', 'bot']
        for arr_ in self.arrs:
            dtype+=[((f'Central time for {arr_} PMTs [ ns ]',
                      f'center_time_{arr_}'),
                      peak_basics_fields['center_time'][0])]
            dtype+=[((f'Time between 10% and 50% area quantiles for {arr_} PMTs [ns]',
                      f'rise_time_{arr_}'),
                      peak_basics_fields['rise_time'][0])]
            dtype+=[((f'Width (in ns) of the central 50% area of the peak for {arr_} PMTs',
                      f'range_50p_area_{arr_}'),
                      peak_basics_fields['range_50p_area'][0])]
            dtype+=[((f'Width (in ns) of the central 90% area of the peak for {arr_} PMTs',
                      f'range_90p_area_{arr_}'),
                      peak_basics_fields['range_90p_area'][0])]
        dtype+=[((f'Difference between center times of top and bottom arrays [ ns ]',
                  f'center_time_diff_top_bot'),
                  peak_basics_fields[f'center_time'][0])]
        dtype += strax.time_fields
        return dtype

    def compute(self, peaks):
        result = np.zeros(peaks.shape, dtype=self.dtype)
        peak_dtype = self.deps['peaks'].dtype
        for arr_ in self.arrs:
            fpeaks_ = np.zeros(peaks.shape[0], dtype=peak_dtype)
            if arr_ == 'top':
                fpeaks_['data']=peaks['data_top']
                fpeaks_['area']=peaks['area']*peaks[f'area_fraction_top']
            elif arr_ == 'bot':
                fpeaks_['data']=(peaks['data']-peaks['data_top'])
                fpeaks_['area']=peaks['area']*(1.-peaks['area_fraction_top'])
            elif arr_ == 'tot':
                # This one is ony
                fpeaks_['data']=peaks[f'{type_}_data']
                fpeaks_['area']=peaks[f'{type_}_area']
            else:
                raise RuntimeError(f'Received unknown array type : '+ arr_)
            fpeaks_['length']=peaks[f'length']
            fpeaks_['dt']=peaks[f'dt']
            mask=(fpeaks_['area']>0)
            # computing center times
            with np.errstate(divide='ignore', invalid='ignore'):
                recalc_ctime = np.sum(fpeaks_['data']*(np.arange(0, fpeaks_['data'].shape[1])), axis=1 )
                recalc_ctime/=fpeaks_['area']
                recalc_ctime*=fpeaks_['dt']
                recalc_ctime[~mask]=0.0
            result[f'center_time_{arr_}']=peaks['time']
            result[f'center_time_{arr_}'][mask]+=recalc_ctime[mask].astype(int)
            # computing widths times
            strax.compute_widths(fpeaks_)
            result[f'rise_time_{arr_}'][:]=np.nan
            result[f'rise_time_{arr_}'][mask]= -fpeaks_['area_decile_from_midpoint'][mask][:, 1]
            result[f'range_50p_area_{arr_}'][:]=np.nan
            result[f'range_50p_area_{arr_}'][mask] = fpeaks_['width'][mask][:, 5]
            result[f'range_90p_area_{arr_}'][:]=np.nan
            result[f'range_90p_area_{arr_}'][mask] = fpeaks_['width'][mask][:, 9]
        result['center_time_diff_top_bot'] = (result[f'center_time_top'] - result[f'center_time_bot'])
        result['time'], result['endtime'] = peaks['time'], strax.endtime(peaks)
        return result



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventInfoDouble(strax.MergeOnlyPlugin):
    """
    Alternate version of event_info for Kr and other double scatter
    analyses:
     - Uses a different naming convention:
       s1 -> s1_a, alt_s1 -> s1_b, and similarly for s2s;
     - Adds s1_b_distinct_channels, which can be tricky to compute
       (since it requires going back to peaks)
    """
    __version__ = '0.1.2'
    depends_on = ['event_info', 'distinct_channels']
    save_when = strax.SaveWhen.EXPLICIT

    @staticmethod
    def rename_field(orig_name):
        special_cases = {'cs1': 'cs1_a',
                         'alt_cs1': 'cs1_b',
                         'alt_s1_delay': 'ds_s1_dt',
                         'cs2': 'cs2_a',
                         'alt_cs2': 'cs2_b',
                         'alt_s2_delay': 'ds_s2_dt',
                         'cs1_wo_timecorr': 'cs1_a_wo_timecorr',
                         'alt_cs1_wo_timecorr': 'cs1_b_wo_timecorr',
                         'cs2_wo_elifecorr': 'cs2_a_wo_elifecorr',
                         'alt_cs2_wo_elifecorr': 'cs2_b_wo_elifecorr',
                         'cs2_wo_timecorr': 'cs2_a_wo_timecorr',
                         'alt_cs2_wo_timecorr': 'cs2_b_wo_timecorr',
                         'cs2_area_fraction_top': 'cs2_a_area_fraction_top',
                         'alt_cs2_area_fraction_top': 'cs2_b_area_fraction_top',
                         'cs2_bottom': 'cs2_a_bottom',
                         'alt_cs2_bottom': 'cs2_b_bottom'}
        if orig_name in special_cases:
            return special_cases[orig_name]

        name = orig_name
        for s_i in [1, 2]:
            if name.startswith(f's{s_i}'):
                name = name.replace(f's{s_i}', f's{s_i}_a')
            if name.startswith(f'alt_s{s_i}'):
                name = name.replace(f'alt_s{s_i}', f's{s_i}_b')
        return name

    def infer_dtype(self):
        self.input_dtype = (
                strax.unpack_dtype(self.deps['event_info'].dtype)
                + [strax.unpack_dtype(self.deps['distinct_channels'].dtype)[0]])
        return [
            ((comment, self.rename_field(name)), dt)
            for (comment, name), dt in self.input_dtype]

    def compute(self, events):
        result = np.zeros(len(events), dtype=self.dtype)
        for (_, name), _ in self.input_dtype:
            result[self.rename_field(name)] = events[name]
        return result
